/* oncSpectrum.c
   Ziheng Yang, 2003 - 2011
   This program estimates mutation rates from p53 database, according to Yang et al. (2003).

      cl -O2 oncSpectrum.c tools.c
      cc -O2 -o oncSpectrum oncSpectrum.c tools.c -lm

              oncSpectrum <datafilename> <refseqfilename>
*/

/*
#define Usep53Version5
*/
#define MAXNP 100

#include "paml.h"

void error2 (char *message);
int Getrefseq(FILE*fout, char* refseqfile, char* intronfile, int *ls, double pi[]);
int CodeCodon(char codon[3]);
char *getcodon(char codon[], int ic);
int ReadMutData(FILE* fout, char*datafile, int ls, int counts[64][64]);
int setparameters(int codonsite, double x[], double *silent, double *kappa, double *mis, double *non, double *CpG, double *diY);
double CodonRate(int nucsite, int targetbase, double silent, double kappa,double mis,double non,double CpG,double diY);
double lfun(double x[], int np);
double lfun_counts(double x[], int np);  /* old & not used now */

FILE *fout, *frub;
extern int NFunCall, noisy, GeneticCode[][64];
extern char BASEs[], AAs[];
extern double Small_Diff;

int icode=0;  /* universal genetic code */
struct MUTDATA {
   int site;    /* nucleotide position in ref sequence */
   char base;   /* target nucleotide after mutation (0123 for TCAG) */
}  *data;

char *refseq=NULL, *refAAseq=NULL, *domains;
char *nucBefore=NULL, *nucAfter=NULL;
double pi[64];
int maxrecords=20000, lseq, nrecords, counts[64][64];
int ndomain, ModelFlag[5], npara[5];

int main(int argc, char*argv[])
{
   char refseqfile[96]="/A/papers/p53/refseq.txt";
#ifdef Usep53Version5
   char datafile[96]="/A/papers/p53/TP53SomaticR5.txt";
   char intronfile[96]="";
#else
   char datafile[96]="/A/papers/p53/TP53SomaticR10.txt";
   char intronfile[96]="/A/papers/p53/TP53SomaticR10.IntronBorders.txt";
#endif
   char outfile[96]="out.txt", timestr[64];
   char *modelstr[5][2]={{"No", "variable"}, {"No", "variable"}, {"No", "variable"},
                         {"No", "Yes"}, {"No", "Yes"}};
   int model, np, i,j,k;
   double x[MAXNP]={2,1,1}, xb[MAXNP][2], lnL, *space, *var;
  
   starttimer();
   printf("Usage:\n\toncSpectrum <datafilename> <refseqfilename> <intronfilename>\nOutput goes to %s\n", outfile);
   if(argc>1) strcpy(datafile, argv[1]);
   if(argc>2) strcpy(refseqfile, argv[2]);
   if(argc>3) strcpy(intronfile, argv[3]);

   fout = (FILE*)gfopen(outfile, "w");
   frub = (FILE*)gfopen("rub","w"); 
   noisy = 3;
   Getrefseq(fout, refseqfile, intronfile, &lseq, pi);

   SetSeed(4*(int)time(NULL)+1, 0);
   /* SetSeed(1, 0); */
   Small_Diff=2e-5;

   np = ndomain - 1 + ndomain + ndomain*2 + 2;
   space = (double*)malloc(max2(spaceming2(np),80000)); var=space+np;
   data = (struct MUTDATA*)malloc(maxrecords*sizeof(struct MUTDATA));
   if(space==NULL || data==NULL) 
      error2("oom space or data");

   ReadMutData(fout, datafile, lseq, counts);
   noisy = 1;

   for(model=0; model<(ndomain>1 ? 32 : 4); model++) {
      for(i=0,k=model; i<5; i++)
         { ModelFlag[5-1-i] = k%2;  k/=2; }
      npara[0] = (ModelFlag[0] ? ndomain-1 : 0);  /* silent */
      npara[1] = (ModelFlag[1] ? ndomain : 1);    /* kappa */
      npara[2] = (ModelFlag[2] ? ndomain*2 : 2);  /* NonMiss */
      npara[3] = ModelFlag[3];                       /* CpG */
      npara[4] = ModelFlag[4];                       /* diY */

      printf("\n\n*** Model ");
      for(i=0; i<5; i++) printf("%d", ModelFlag[i]);
      printf(" ***\n%10s%10s%10s%10s%10s%10s\n", "", "silent", "kappa", "NonMiss", "CpG", "diY");
      printf("%10s", "");
      for(i=0; i<5; i++) printf("%10s", modelstr[i][ModelFlag[i]]);
      printf("\nnpara    ");
      for(i=0; i<5; i++) printf("%10d", npara[i]);
      printf("\nnp = %d", np = npara[0]+npara[1]+npara[2]+npara[3]+npara[4]);

      fprintf(fout, "\n\n*** Model ");
      for(i=0; i<5; i++) fprintf(fout, "%d", ModelFlag[i]);
      fprintf(fout, "\n%10s%10s%10s%10s%10s%10s\n", "", "silent", "kappa", "NonMiss", "CpG", "diY");
      fprintf(fout, "%10s", "");
      for(i=0; i<5; i++) fprintf(fout, "%10s", modelstr[i][ModelFlag[i]]);
      fprintf(fout, "\nnpara    ");
      for(i=0; i<5; i++) fprintf(fout, "%10d", npara[i]);
      fprintf(fout, "\n\nnp = %d", np = npara[0]+npara[1]+npara[2]+npara[3]+npara[4]);

      fprintf(frub, "\n\n*** Model ");
      for(i=0; i<5; i++) fprintf(frub, "%d", ModelFlag[i]);

      for(i=0; i<np; i++) x[i]=.1+rndu()*5;  /* random initial values */
      for(i=0; i<np; i++) { xb[i][0]=.005; xb[i][1]=995; }  /* lower and upper bounds */
      lnL = lfun(x,np);
      if(noisy>1) {
         matout(F0,x,1,np);
         printf("\nlnL0 = %12.6f\n",-lnL);
      }

      j = ming2(frub, &lnL, lfun, NULL, x, xb, space, 1e-9, np);
      Hessian (np, x, lnL, space, var, lfun, var+np*np);
      matinv (var, np, np, var+np*np);
      for(i=0; i<np; i++) var[i*np+i] = (var[i*np+i]>0. ? sqrt(var[i*np+i]) : -1);

      printf("\nlnL:%14.6f\nM.L.E.s:\n", -lnL);
      for(i=0; i<np; i++) printf(" %8.5f",x[i]);
      printf ("\nS.E.s\n");
      for(i=0; i<np; i++) printf(" %8.5f", var[i*np+i]);  FPN(F0);

      fprintf (fout,"\nlnL:%14.6f\nMLEs:\n", -lnL);
      for(i=0; i<np; i++) fprintf(fout," %8.5f",x[i]);
      fprintf (fout,"\nSE\n");
      for(i=0; i<np; i++) fprintf(fout," %8.5f", var[i*np+i]);  FPN(fout);
      fflush(fout);
      printf("\nTime used: %s\n", printtime(timestr));
   }
   free(space);
   free(data);
   free(refseq);
   if(nucBefore) free(nucBefore);
         
   return(0);
}

int CodeCodon(char codon[])
{
/* code a codon (e.g. TCA) into 0, 1, ..., 63
*/
   int i, ic, b;

   for(i=0,ic=0; i<3; i++)  {
      b=toupper(codon[i]);
      switch (b) {
         case 'T':  case 'U':   b=0; break;
         case 'C':              b=1; break;
         case 'A':              b=2; break;
         case 'G':              b=3; break;
         default: error2("ambiguity character in reference sequence");
      }
      ic=ic*4+b;
   }
   return(ic);
}


double Fcodon_3x4(double fcodon[], double fb3x4[]);
void OutFb3x4 (FILE*fout, double fb3x4[]);

double Fcodon_3x4(double fcodon[], double fb3x4[])
{
/* this converts the codon frequencies into a fb3x4 table. fcodon has 64 codons.
*/
   int b[3], k,j, nc=64;
   double t;

   zero(fb3x4,12);
   for(k=0; k<nc; k++) {
      b[0]=k/16; b[1]=(k%16)/4; b[2]=k%4;
      for(j=0; j<3; j++)
         fb3x4[j*4+b[j]] += fcodon[k];
   }
   for(j=0; j<3; j++) {
      t = sum(fb3x4+j*4, 4);
      if(t<1e-15) error2("Fcodon_3x4: empty sequence?");
      abyx(1/t, fb3x4+j*4, 4);
   }
   return(0);
}

void OutFb3x4 (FILE*fout, double fb3x4[])
{
   int j,k;
   for(j=0; j<3; j++) {
      fprintf(fout, "\nposition %2d:", j+1);
      for(k=0; k<4; k++)
         fprintf(fout, "%5c:%7.5f", BASEs[k], fb3x4[j*4+k]);
   }
}


int Getrefseq(FILE*fout, char* refseqfile, char* intronfile, int *lseq, double pi[])
{
/* This reads the reference sequence into refseq[], and code bases TCAG into 0123.
   *lseq is the number of nucleotides in the refseq (3 times the number of codons).
   This also reads the site (codon) labels to partition the gene into different domains.
   It reads the file of intron boundaries and sets up nucBefore and nucAfter.
*/
   FILE *fin;
   int lline=16000, h, ch, i;
   int maxnintron=900, nintron=0, loc[900], bstart[900], bend[900], intron;
   char line[16000+1], goodlabel, bases[6]="TCAG-";
   int debug=0;

   printf("Reading reference sequence from %s\n", refseqfile);
   fin = gfopen(refseqfile,"r");
   fscanf(fin, "%d", lseq);
   refseq = (char*)malloc(*lseq * 2 * sizeof(char));
   refAAseq = refseq + *lseq;
   domains = refAAseq + *lseq/3;
   for(i=0; i< *lseq/3; i++) domains[i]=0;
   if(refseq==NULL) error2("oom refseq");

   for(h=0; h< *lseq; ) {
      ch = fgetc(fin);
      if(ch==EOF)     error2("EOF in refseq file"); 
      if(!isalpha(ch)) continue;
      refseq[h++] = CodeChara((char)ch,0);
      if(h == *lseq) break;
   }
   for(h=0; h< *lseq/3; h++) {
      ch = refseq[h*3]*16+refseq[h*3+1]*4+refseq[h*3+2];
      refAAseq[h] = GeneticCode[icode][ch];
      if(refAAseq[h]==-1) refAAseq[h] = 20;
      pi[ch]++;
   }

   /* try to read site labels to use partition models. Labels must be consecutive */
   fgets(line, lline, fin);
   for(h=0,ndomain=0; h< *lseq/3; h++) {
      for( ; ; ) {
         ch=fgetc(fin);
         if(isdigit(ch)) { goodlabel=1; ch=ch-(int)'1'; break; }
         else if (ch==EOF || isgraph(ch)) { goodlabel=0; break; }
      }
      if(goodlabel) {
         domains[h] = ch;
         ndomain = max2(ndomain,ch+1);
      }
      else 
         break;
   }
   if(h!= *lseq/3) {  /* have not read enough labels */
      printf("The whole gene is analyzed as one partition.\n");
      ndomain = 1;
   }
   else {
      printf("%d partitions (domains) are assumed.  Site labels are \n", ndomain);
      for(i=0; i< *lseq/3; i++)
         printf("%d", domains[i]+1);
      FPN(F0);
   }
   fclose(fin);
   if(intronfile[0]=='\0') return(0);
   
   printf("Reading intron info from %s\n", intronfile);
   nucBefore = (char*)malloc(*lseq * 2 * sizeof(char));
   nucAfter = nucBefore + *lseq;
   for(h=1,nucBefore[0]=4; h < *lseq; h++)
      nucBefore[h] = refseq[h-1];
   for(h=0,nucAfter[*lseq-1]=4; h < *lseq-1; h++)
      nucAfter[h] = refseq[h+1];
   fin = gfopen(intronfile,"r");
   fscanf(fin, "%d", &nintron);
   if(nintron<1) { nintron=0; printf("nintron<1.  intronBorder file not used."); }
   fgets(line, lline, fin);
   for(i=0; i<nintron; i++) {
      fscanf(fin, "%d %c..%c", &loc[i], &bstart[i], &bend[i]);
      fgets(line, lline, fin);
      printf("  intron %2d loc %5d  %c..%c\n", i+1, loc[i], bstart[i], bend[i]);

      bstart[i] = strchr(bases, bstart[i]) - bases;
      bend[i]   = strchr(bases, bend[i]) - bases;
      if(bstart[i]<0 || bstart[i]>3 || bend[i]<0 || bend[i]>3)
         error2("start or end nucleotide strange");
      nucAfter [loc[i]-1 - 1] = bstart[i];
      nucBefore[loc[i]-1] = bend[i];
   }
   /* print refseq, before & after for error checking. */
   if(debug) {
      printf("\nNuc  Before.Ref.After\n");
      for(h=0, intron=0; h < *lseq; h++) {
         if(h+1 == loc[intron]) {
            printf("   * intron %2d: %c..%c", intron+1, tolower(bases[bstart[intron]]), tolower(bases[bend[intron]]) );
            if(++intron > nintron) error2("strange");
         }
         printf("\n%5d %4c . %c . %c", h+1, bases[nucBefore[h]], bases[refseq[h]], bases[nucAfter[h]]);
         if((h+1)%30 == 0) getchar();
      }
   }
   return(0);
}

int LineToFields(char *line, char fields[][64], char separator, 
                 int lline, int lfield, int field2read)
{
/* this separates a line into fields, separated by separator
   lline is the max line length, lfield is max field length, 
   and maxfield is max number of fields needed; the rest is discarded

  A zero is attached to the end of each field string.
  Not carefully written. So allow one extra for lline, lfield, maxfield 
  etc.
*/
   int i, ifield=0, charread=0;
   char *p=line;

   for(i=0; i<lline && *p!='\n' && *p && ifield<field2read; i++,p++) {
      if(*p == separator)
         { fields[ifield][charread]='\0';  ifield++;  charread=0; }
      else 
         fields[ifield][charread++] = *p;
      if(charread>=lfield) 
         error2("raise field length");
   }
   fields[ifield][charread]=0;
   return(0);
}

int ReadMutData(FILE* fout, char*datafile, int lseq, int counts[64][64])
{
/* this reads and processes the mutation data file line by line.  
   The results are in a mutation count matrix counts[64][64] (used 
   in lfun_counts), and the vector MutData (used in lfun).
*/
   /* change the index of the following fields according to the database.  Index starts from 0. */
#ifdef Usep53Version5
   enum {Type=1, ExonIntron=2, CodonNumber=3, WT_codon=6, Mutant_codon=7, WT_AA=8, Mutant_AA=9};
#else
   enum {Type=6, ExonIntron=1, CodonNumber=2, WT_codon=8, Mutant_codon=9, WT_AA=10, Mutant_AA=11};
#endif
   int  lline=32000, field2read=12, lfield=64;
   char line[32000+1], fields[100][64], separator='\t';
   FILE *fin, *fused;
   char *codons[2]={"   ","   "}, AA3[4]="   ", searchstr[32];
   int n=64, lc=lseq/3, recNO, nread, i,j, location, pos=0, ic[2], iaa[2];
   /* counts at sites and sums */
   int *ins, *del, *diff23, *silent, *replace,*stop, intron=0; 
   int filterdata=0;

   if(filterdata) {
      puts("Search string to filter data");
      scanf("%s", searchstr);
      if(!isalpha(searchstr[0])) filterdata=0;
      else                       strcase(searchstr, 1);
   }
   if(filterdata && (fused=(FILE*)fopen("RECused","w"))==NULL)  
      error2("can't open RECused");

   printf("\nReading data from %s\n\n", datafile);
   if((ins=(int*)malloc(6*(lc+1)*sizeof(int)))==NULL) error2("oom ReadMutData");
   del = ins+lc+1;
   diff23 = del+lc+1; 
   silent = diff23+lc+1;
   replace = silent+lc+1;
   stop = replace+lc+1;
   for(i=0; i<6*(lc+1); i++) ins[i]=0;
   for(i=0; i<n; i++) for(j=0; j<n; j++) counts[i][j]=0;

   fin=gfopen(datafile,"r");
   fgets(line, lline, fin);  /* remove first line, header */
   for(nread=nrecords=0; ; nread++) {
      if(fgets(line, lline, fin) == NULL) break;
      LineToFields(line, fields, separator, lline, lfield, field2read);

      if(filterdata && strstr(line,searchstr)==NULL) continue;

      if(strstr(fields[ExonIntron],"intron")) { intron++; continue; }

      sscanf(fields[0], "%d", &recNO);
      sscanf(fields[CodonNumber], "%d", &location);
      if(location>lc) error2("location too large");
      if(location==0)
         { puts("location = 0?"); continue; }
      location--;

      if(strstr(fields[Type],"del")) { del[location]++; del[lc]++; continue; }
      else if(strstr(fields[Type],"ins")) { ins[location]++; ins[lc]++; continue; }
      else if(!strstr(fields[Type],">"))  continue;


      if(noisy) printf("\r%-5d recNO %6d (codon %3d) %s (%s) -> %s (%s)   ", 
                       nread+1, recNO, location+1, fields[WT_codon],
                       fields[WT_AA], fields[Mutant_codon],fields[Mutant_AA]);

      for(i=0,j=0,pos=-1;i<3;i++) 
         if(fields[WT_codon][i]!=fields[Mutant_codon][i]) { j++; pos=i; }
      if(j<1) { printf("same source and target!\n"); continue; }
      else if (j>1) {
         printf(" %d differences\n", j);
         diff23[location]++;  diff23[lc]++;
         continue; 
      }
      ic[0]=CodeCodon(fields[WT_codon]);      iaa[0]=GeneticCode[icode][ic[0]];
      ic[1]=CodeCodon(fields[Mutant_codon]);  iaa[1]=GeneticCode[icode][ic[1]];
      j = refseq[location*3]*16+refseq[location*3+1]*4+refseq[location*3+2];
      if(j != ic[0]) error2("\ncodon does not match ref seq\n");
      
      for(i=0;i<2;i++) {
         if(i==1 && iaa[1]==-1) {
            if(!strstr(fields[Mutant_AA],"STOP")) 
               error2("should be STOP");
         }
         else {
            getAAstr(AA3,iaa[i]);
            if(!strstr(fields[WT_AA+i],AA3)) puts("different AAs");
         }
      }
      if(iaa[0]==-1)          error2("source codon is a stop codon");
      else if(iaa[1]==-1)     { stop[location]++; stop[lc]++; }
      else if(iaa[0]==iaa[1]) { silent[location]++; silent[lc]++; }
      else if(iaa[0]!=iaa[1]) { replace[location]++; replace[lc]++; }

      counts[ic[0]][ic[1]] ++;
      data[nrecords].site=location*3+pos;
      data[nrecords].base=CodeChara(fields[Mutant_codon][pos],0);
      if(++nrecords>maxrecords) error2("raise maxrecords & recompile");
      if(filterdata)
         fprintf(fused,"%d\t%s\t%s\t%s\n", nrecords, fields[0],fields[1],fields[2]);

   }  /* for(nread=nrecords=0) */
/*
   fprintf(fout,"\nCounts of source and target codon pairs\n");
   for(i=0; i<n; i++,FPN(fout)) for(j=0; j<n; j++)
      fprintf(fout," %3d", counts[i][j]);
*/
   fprintf(fout,"\n\n%d records in data file, %d used\n", nread, nrecords);
      fprintf(fout,"%d ins, %d dels, %d introns, %d diff23, %d silent, %d replacement, %d stops\n", 
         ins[lc], del[lc],intron,diff23[lc],silent[lc],replace[lc],stop[lc]);
      fprintf(fout,"Site\t\tIns\tDel\tDiff23\tsilent\treplace\tstop\n");
      for(i=0; i<lc; i++,FPN(fout)) {
         fprintf(fout,"%d\t%c%c%c %c", i+1, BASEs[refseq[i*3]],BASEs[refseq[i*3+1]],BASEs[refseq[i*3+2]],AAs[refAAseq[i]]);
         for(j=0; j<6; j++)  fprintf(fout,"\t%d", ins[j*(lc+1)+i]);
      }
   if(noisy) {
      printf("\n\n%d records in data file, %d used\n", nread, nrecords);
      printf("%d ins, %d dels, %d introns, %d diff23, %d silent, %d replacement, %d stops\n", 
         ins[lc], del[lc],intron,diff23[lc],silent[lc],replace[lc],stop[lc]);
   }
   free(ins);
   fclose(fin);
   if(filterdata) fclose(fused);
   return(0);
}



double lfun_counts (double x[], int np)
{
/* This calculates the likelihood function using counts[][] of mutations.  
   It constructs the rate matrix Q first.  It is now superceded by lfun, which
   deals with CpG doublet rates by looking at quintets/triplets along the sequence.

   Definition of variables.
      mutfreqs[i]: probability that the mutated codon in the refseq is codon i.
      ndiff: number of differences between codons i and j.  The rate Q[i][j] 
      is 0 if the two codons are different at 2 or 3 positions (ndiff=2 or 3). 
      The single difference is at codon position pos.
      aa1 and aa2 are amino acids coded by codons i and j.
      from[3] are the three bases of the source codon i, and to[3] are the 
      three bases of the target codon j.  Bases TCAG are coded 0123, and codons 
      are ordered TTT, TTC, ..., GGG.  So for source codon TCA, we have i = 6 
      and from[0]=0; from[1]=1; from[2]=2;
   Globals used:
      pi[i] is the frequency of codon i in refseq
      counts[i][j] is the number of i->j substitutions in the mutation data.
*/
   double kappa, missense, nonsense;
   double Q[64][64], mutfreqs[64], sum_mutfreqs=0;
   double y, lnL;
   int n=64, from[3],to[3], i,j,k, ndiff, pos, aa1,aa2;

   NFunCall++;
   kappa=x[0], missense=x[1], nonsense=x[2];
   for (i=0; i<n; i++) for (j=0; j<n; j++) Q[i][j]=0;
   for (i=0; i<n; i++) {
      aa1=GeneticCode[icode][i];
      for (j=0; j<n; j++) {
         from[0]=i/16; from[1]=(i/4)%4; from[2]=i%4; /* three bases of codon i */
         to[0]=j/16;   to[1]=(j/4)%4;   to[2]=j%4;   /* three bases of codon j */
         for (k=0,ndiff=0; k<3; k++)  
            if (from[k]!=to[k]) { ndiff++; pos=k; }
         if (ndiff!=1) continue;  /* rate is nonzero only if one difference */
         if(from[pos]+to[pos]==1 || from[pos]+to[pos]==5)  /* transition */
            Q[i][j]=kappa;
         else                                              /* transversion */
            Q[i][j]=1;
         aa2 = GeneticCode[icode][j];

         if(aa2==-1)           Q[i][j] *= nonsense;   /* stop */
         else if(aa1!=aa2)     Q[i][j] *= missense;   /* missense */
      }
      for (j=0,y=0; j<n; j++)  y += Q[i][j];
      for (j=0; j<n; j++) Q[i][j] /= y;
      mutfreqs[i]=pi[i]*y;
      sum_mutfreqs+=mutfreqs[i];
   }
   for (i=0; i<n; i++)  mutfreqs[i] /= sum_mutfreqs;

   for (i=0,lnL=0; i<n; i++) {
      for (j=0; j<n; j++) {
         if(counts[i][j]) {
            if(Q[i][j]==0) 
               error2("P=0");
            lnL -= counts[i][j]*log(mutfreqs[i]*Q[i][j]);
         }
      }
   }
   return(lnL);
}


double CodonRate(int nucsite, int targetbase, double silent, double kappa,double mis,double non,double CpG,double diY)
{
/* this calculates q(nucsite, targetbase), the rate that the base at nucsite 
   changes into targetbase.  h is the codon site (nucsite/3), and pos is the 
   codon position (1,2,3).  
   codon[2] are the source and target codons, triplet[2] are the triplets, aa[2] 
   are the amino acids.
*/
   int h=nucsite/3, pos=nucsite%3+1, k, CG1,CG2, diY1,diY2, T=0,C=1,A=2,G=3;
   char codon[2][3], triplet[2][3], aa[2];
   double q;

   for(k=0; k<3; k++)
      codon[0][k] = codon[1][k] = refseq[h*3+k];
   triplet[0][0] = triplet[1][0] = (nucsite==0 ? -1 : refseq[h*3+pos-1-1]);
   triplet[0][1] = triplet[1][1] = refseq[h*3+pos-1];
   triplet[0][2] = triplet[1][2] = (nucsite==lseq-1 ? -1 : refseq[h*3+pos-1+1]);
   if(nucBefore) {
      triplet[0][0] = triplet[1][0] = nucBefore[nucsite];
      triplet[0][2] = triplet[1][2] = nucAfter[nucsite];
   }
   codon[1][pos-1] = triplet[1][1] = targetbase;
   aa[0] = refAAseq[h];
   aa[1] = GeneticCode[icode][codon[1][0]*16+codon[1][1]*4+codon[1][2]];

   q = silent;
   k = codon[0][pos-1] + codon[1][pos-1];
   if(k==1 || k==5) q *= kappa;
   if(aa[1]==-1)                    q *= non;   /* stop */
   else if(aa[0]!=aa[1])            q *= mis;   /* missense */
   for(k=0,CG1=0; k<2; k++)  if(triplet[0][k]==C && triplet[0][k+1]==G) CG1++;
   for(k=0,CG2=0; k<2; k++)  if(triplet[1][k]==C && triplet[1][k+1]==G) CG2++;
   if(CG1 && !CG2)       q *= CpG;      /* from CpG */
   else if(!CG1 && CG2)  q /= CpG;      /*   to CpG */

   for(k=0,diY1=0; k<2; k++)  
      if((triplet[0][k]==T||triplet[0][k]==C) && (triplet[0][k+1]==T||triplet[0][k+1]==C)) diY1++;
   for(k=0,diY2=0; k<2; k++)  
      if((triplet[1][k]==T||triplet[1][k]==C) && (triplet[1][k+1]==T||triplet[1][k+1]==C)) diY2++;
   if(diY1 && !diY2)       q *= diY;      /* from di-pyrimidine */
   else if(!diY1 && diY2)  q /= diY;      /*   to di-pyrimidine */

   return (q);
}


int setparameters(int codonsite, double x[], double *silent, double *kappa, 
                  double *mis, double *non, double *CpG, double *diY)
{
   int k, domain = domains[codonsite];  /* domain labels: 0, 1, 2, ..., ndomain-1 */

   *silent = (ModelFlag[0]&&domain ? x[domain-1] : 1);
   *kappa = x[ npara[0] + (ModelFlag[1] ? domain : 0) ];
   *non = x[npara[0] + npara[1] + (ModelFlag[2] ? domain : 0)]; 
   *mis = x[npara[0] + npara[1] + (ModelFlag[2] ? ndomain+domain : 1)];
   k = npara[0] + npara[1] + npara[2];
   *CpG = *diY = 1;
   if(ModelFlag[3]) *CpG = x[k++];
   if(ModelFlag[4]) *diY = x[k++];
   if(k!=npara[0]+npara[1]+npara[2]+npara[3]+npara[4])
      error2("np error");
   return(k);
}

double lfun (double x[], int np)
{
/* This calculates the likelihood function by going through the records.  
   It first goes through codons in the ref seq to calculate the total rate sumr.
*/
   double q, silent=1, kappa, non, mis, CpG, diY;
   double lnL, sumr=0;
   int n=64, i,j, h, pos, part=0;

   NFunCall++;
   for (h=0; h<lseq/3; h++) { /* calculates total rate sumr along the ref seq */
      setparameters(h, x, &silent, &kappa, &non, &mis, &CpG, &diY);
      for(pos=0; pos<3; pos++) /* cycles through 9 neighbors */
         for(j=0;j<4;j++)
            if(j!=refseq[h*3+pos]) 
               sumr += CodonRate(h*3+pos, j, silent, kappa, non, mis, CpG, diY);
   }
   for (i=0,lnL=0; i<nrecords; i++) { /* sum over nrecords to calculate lnL */
      setparameters(data[i].site/3, x, &silent, &kappa, &non, &mis, &CpG, &diY);
      q = CodonRate(data[i].site, data[i].base, silent, kappa, non, mis, CpG, diY);
      lnL -= log(q/sumr);
   }
   return (lnL);
}
